"""
    Drissionpage-mcp的基础功能
"""
import hashlib
import json
import os
import time
from typing import Any

from DrissionPage.errors import NoRectError
from fastmcp import FastMCP

from tools.browser_manager import BrowserManager
from tools.browser_proxy import DPProxyClientManager
from tools.tools import dp_mcp_message_pack, compress_html_js, compress_image_bytes


def register_visit_url(mcp: FastMCP, browser_manager: BrowserManager, client_manager: DPProxyClientManager):
    @mcp.tool(name="visit_url",
              description="使用Drissionpage打开url访问某个网站，并开始监听初始tab页的所有的XHR请求"
                          "当需要使用手机版浏览器Ua时use_mobile_user_agent为True"
                          "如果想要以域名对packet进行过滤，可以传入想要过滤的域名列表。默认是：None。"
                          "如果想要以method对packet进行过滤，可以传入想要过滤的method列表，默认是：['GET', 'POST']")
    async def visit_url(url: str, domain_filter: list = None, method_filter: list = ["GET", "POST"],
                        use_mobile_user_agent: bool = False) -> dict[str, Any]:
        mobile_user_agent = None
        if use_mobile_user_agent:
            mobile_user_agent = "Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/143.0.0.0 Mobile Safari/537.36"
        port, _browser = browser_manager.create_browser(mobile_user_agent)
        tab = _browser.get_tab()
        packet_filter = {
            "domain_filter": domain_filter,
            "method_filter": method_filter,
        }
        client_manager.create_client(tab, packet_filter)
        tab.get(url)
        tab_id = tab.tab_id
        return dp_mcp_message_pack(
            f"已在[{port}]端口创建浏览器对象，并已打开链接：{url}，打开的模式是：{'手机版' if use_mobile_user_agent else '电脑版'}",
            tab_id=tab_id,
            browser_port=port
        )


def register_get_new_tab(mcp: FastMCP, browser_manager, client_manager: DPProxyClientManager):
    @mcp.tool(name="get_new_tab",
              description="使用Drissionpage创建一个新的tab页，在新的tab页中打开url，并开始监听新的tab页的所有XHR请求"
                          "如果想要以域名对packet进行过滤，可以传入想要过滤的域名列表。默认是：None。"
                          "如果想要以method对packet进行过滤，可以传入想要过滤的method列表，默认是：['GET', 'POST']")
    async def get_new_tab(browser_port: int, url: str, domain_filter: list = None,
                          method_filter: list = ["GET", "POST"]) -> dict[str, Any]:
        _browser = browser_manager.get_browser(browser_port)
        tab = _browser.new_tab()
        packet_filter = {
            "domain_filter": domain_filter,
            "method_filter": method_filter,
        }
        client_manager.create_client(tab, packet_filter)
        tab.get(url)
        _browser.activate_tab(tab)
        tab_id = tab.tab_id
        return dp_mcp_message_pack(f"已创建新的tab页，并打开链接：{url}", tab_id=tab_id)


def register_pop_first_packet(mcp: FastMCP, browser_manager, client_manager: DPProxyClientManager):
    @mcp.tool(name="pop_first_packet",
              description="每调用一次就会弹出传入的tab页所监听到的数据包中的第一个packet_message，当一个packet_message的response body过长时会被切分成多个包，具体一个请求是否还有下一个包，可以参考body_completed字段")
    async def pop_first_packet(browser_port: int, tab_id: str) -> dict[str, Any]:
        _browser = browser_manager.get_browser(browser_port)
        client = client_manager.get_client(tab_id)
        current_queue_size, packet_message = client.pop_first_packet()
        message = f"tab页:【{tab_id}】，暂时没有监听到XHR数据包"
        if packet_message:
            message = f"tab页:【{tab_id}】，监听到XHR数据包，当前数据包队列中还剩 {current_queue_size} 条数据，如果还剩数据为0，可以暂时稍后再次调用该方法"
        if (packet_message is None) and current_queue_size:
            message = f"tab页:【{tab_id}】，当前弹出的第一个数据包不符合过滤条件，当前数据包队列中还剩 {current_queue_size} 条数据，请不要改变条件，继续弹出下一个数据包"
        return dp_mcp_message_pack(
            message,
            browser_port=browser_port,
            tab_id=tab_id,
            packet_message=packet_message,
            current_queue_size=current_queue_size,
        )


# todo：这个获取html的方式后续需要修改，后面采用的思路应该是：获取全部html，然后提取一个高度抽象化的html结构化对象，然后提供一系列的工具对这个对象进行查询，对象至少包含如下属性和工具
#  属性：
#   1. id【用于通过id查询特定的html元素】
#  工具：
#   1. 根据id查询特定的html元素
#   2. 根据传入的正则表达式去html全文检索，如果匹配到则返回所有匹配到的元素id【最好可以给出除id以外的其他更准确的信息，否则ai在查询出来一大堆元素后需要逐个元素调用1.的方法会大幅拖慢处理速度】
def register_get_html(mcp: FastMCP, browser_manager):
    @mcp.tool(name="get_html", description="使用Drissionpage获取某一个tab页的html")
    async def get_html(browser_port: int, tab_id: str) -> dict[str, Any]:
        _browser = browser_manager.get_browser(browser_port)
        tab = _browser.get_tab(tab_id)
        file_name_prefix = hashlib.md5(str(tab.title).encode('utf-8')).hexdigest()

        # min_html, compress_rate = compress_html(tab.html)
        min_html = tab.run_js(compress_html_js)
        # html_str_list = [min_html[i:i + one_turn_max_token] for i in range(0, len(min_html), one_turn_max_token)]
        html_file_list = []
        for index, html_str in enumerate([min_html]):
            file_name = file_name_prefix + f"_{tab_id}_segment{index}.html"
            abs_path = os.path.join(browser_manager.get_cache_path(), file_name)
            with open(abs_path, "w", encoding="utf-8") as f:
                f.write(html_str)
            html_file_list.append(abs_path)
        message = f"已保存tab页：【{tab_id}】的html源码片段共{len(html_file_list)}个"
        return dp_mcp_message_pack(message, tab_id=tab_id, htmls_local_path=html_file_list)


def register_switch_tab(mcp: FastMCP, browser_manager):
    @mcp.tool(name="switch_tab", description="根据传入的tab_id切换到对应的tab页", )
    async def switch_tab(browser_port: int, tab_id: str) -> dict[str, Any]:
        _browser = browser_manager.get_browser(browser_port)
        _browser.activate_tab(tab_id)
        return dp_mcp_message_pack(f"已将tab页:【{tab_id}】切换至最前端")


def register_close_tab(mcp: FastMCP, browser_manager):
    @mcp.tool(name="close_tab", description="根据传入的tab_id关闭tab页", )
    async def close_tab(browser_port: int, tab_id: str) -> dict[str, Any]:
        _browser = browser_manager.get_browser(browser_port)
        _browser.close_tabs(tab_id)
        return dp_mcp_message_pack(f"已将tab页:【{tab_id}】关闭")


def register_check_selector(mcp: FastMCP, browser_manager):
    @mcp.tool(name="check_selector",
              description="查找tab页中是否包含元素，并以列表的形式返回元素信息"
                          "当要选择的元素包含过多元素时，需要传入offset和page_size来分批查看元素，一般不建议调整page_size，更推荐你调整offset"
                          "同时如果单个元素属性值太长，函数会进行截断。一般的单个元素的属性值超过100个字符的就会触发截断，截断后会在最后拼接'...'")
    async def check_selector(browser_port: int, tab_id: str, css_selector: str, attr_name: str = "text",
                             offset: int = 0, page_size: int = 10) -> dict[str, Any]:
        _browser = browser_manager.get_browser(browser_port)
        target_tab = _browser.get_tab(tab_id)
        slice_seg_len = 100
        exist_flag = False
        output_list = []
        if "css:" not in css_selector:
            css_selector = "css:" + css_selector
        if "css:*" == css_selector:
            return dp_mcp_message_pack(
                message="禁止使用全局匹配的Selector：*，请重新生成更精确的Selector",
                tab_id=tab_id,
                selector=css_selector,
            )
        target_eles = target_tab.eles(css_selector)
        total_ele_len = len(target_eles)
        if len(target_eles) != 0:
            exist_flag = True
        if len(target_eles) > page_size:
            target_eles = target_eles[offset:offset + page_size]
        for index, target_ele in enumerate(target_eles):
            temp_dict = {}
            attr_value = "".join(target_ele.texts()).replace("\n", "")
            if attr_name != "text":
                attr_value = target_ele.attr(attr_name) if target_ele.attr(attr_name) else ""
            is_truncate = len(attr_value) > slice_seg_len
            attr_value = attr_value[:slice_seg_len]
            if is_truncate:
                attr_value += "..."
            temp_dict["element_index"] = offset + index
            temp_dict["attr_name"] = attr_name
            temp_dict["attr_value"] = attr_value
            temp_dict["is_truncate"] = is_truncate
            try:
                has_rect = target_ele.states.has_rect
                is_covered = target_ele.states.is_covered
            except NoRectError:
                has_rect = False
                is_covered = "元素没有大小和位置信息"
            temp_dict["ele_info"] = {
                "元素tag": target_ele.tag,
                "元素是否拥有大小和位置信息": has_rect,
                "元素是否可被模拟点击": target_ele.states.is_clickable,
                "元素是否可见": target_ele.states.is_displayed,
                "元素是否可用": target_ele.states.is_enabled,
                "元素是否被其它元素覆盖": is_covered,
            }
            output_list.append(temp_dict)

        return dp_mcp_message_pack(
            f"已完成tab页:【{tab_id}】对：【{css_selector}】的检查，Selector共选中了 {total_ele_len} 个元素",
            tab_id=tab_id,
            selector=css_selector,
            selector_ele_exist=exist_flag,
            page_size=page_size,
            offset=offset,
            ele_info=output_list,
        )


def register_quit_browser(mcp: FastMCP, browser_manager):
    @mcp.tool(name="quit_browser", description="退出浏览器会话，关闭浏览器")
    async def quit_browser(browser_port: int) -> dict[str, Any]:
        flag, _browser = browser_manager.remove_page(browser_port)
        if flag:
            _browser.quit()
        return dp_mcp_message_pack(
            f"浏览器[{browser_port}]，退出会话，关闭浏览器{'成功' if flag else '失败'}",
            browser_port=browser_port,
            quit_flag=flag
        )


def register_get_ele_info(mcp: FastMCP, browser_manager):
    @mcp.tool(name="get_ele_info", description="返回元素有关一系列的信息。"
                                               "参数说明："
                                               "browser_port：浏览器端口号。"
                                               "tab_id：tab页的id。"
                                               "css_selector：目标选择器"
                                               "element_index：如果选择器定位到的是一个列表【或者可以定位到多个元素】，则该值用于定位这个列表中具体元素，默认值为0【列表中的第一个元素】"
                                               "返回值说明："
                                               "element_tag：此属性返回元素的标签名。"
                                               "element_attrs_key：此属性以list的形式返回元素所有属性的key。"
                                               "element_rect_size：此属性以元组形式返回元素的大小【如果元素没有位置及大小，则返回空元组】。"
                                               "is_in_viewport：此属性以布尔值方式返回元素是否在视口中，以元素可以接受点击的点为判断。"
                                               "is_whole_in_viewport：此属性以布尔值方式返回元素是否整个在视口中。"
                                               "is_alive：此属性以布尔值形式返回当前元素是否仍可用。用于判断是否因页面刷新而导致元素失效。"
                                               "is_checked：此属性以布尔值返回表单单选或多选元素是否选中。"
                                               "is_selected：此属性以布尔值返回<select>元素中的项是否选中。"
                                               "is_enabled：此属性以布尔值返回元素是否可用。"
                                               "is_displayed：此属性以布尔值返回元素是否可见。"
                                               "is_covered：此属性返回元素是否被其它元素覆盖/遮挡。如被覆盖/遮挡，返回覆盖元素的 id【无障碍树的id，无法用于Selector】，否则返回False。"
                                               "is_clickable：此属性返回元素是否可被模拟点击，从是否有大小、是否可用、是否显示、是否响应点击判断，不判断是否被遮挡。"
                                               "has_rect：此属性返回元素是否拥有大小和位置信息，有则返回四个角在页面上的坐标组成的列表，没有则返回False。")
    async def get_ele_info(browser_port: int, tab_id: str, css_selector: str, element_index: int = 0) -> dict[
        str, Any]:
        _browser = browser_manager.get_browser(browser_port)
        target_tab = _browser.get_tab(tab_id)
        css_selector = css_selector
        if "css:" not in css_selector:
            css_selector = "css:" + css_selector
        target_eles = target_tab.eles(css_selector)
        try:
            target_element = target_eles[element_index]
        except IndexError:
            return dp_mcp_message_pack(
                message="报错：IndexError: list index out of range。请检查Selector中是否包含了如：first-child、nth-child等字段，如果有则去掉。必须使用element_index来控制元素的选择",
                browser_port=browser_port,
                tab_id=tab_id,
                css_selector=css_selector,
                element_index=element_index,
            )

        try:
            has_rect = target_element.states.has_rect
            is_covered = target_element.states.is_covered
            element_rect_size = target_element.rect.size
            is_in_viewport = target_element.states.is_in_viewport
            is_whole_in_viewport = target_element.states.is_whole_in_viewport
            rect_x_center = (has_rect[0][0] + has_rect[1][0]) / 2
            rect_y_center = (has_rect[1][1] + has_rect[2][1]) / 2
            rect_center = (rect_x_center, rect_y_center)
        except NoRectError:
            has_rect = False
            is_covered = "元素没有大小和位置信息"
            element_rect_size = "元素没有大小和位置信息"
            rect_center = "元素没有大小和位置信息"
            is_in_viewport = "元素没有大小和位置信息"
            is_whole_in_viewport = "元素没有大小和位置信息"
        try:
            child_count = target_element.child_count
        except Exception:
            child_count = 0
        return dp_mcp_message_pack(
            message="元素可以被正常的选择到，以下是元素的信息",
            browser_port=browser_port,
            tab_id=tab_id,
            css_selector=css_selector,
            element_index=element_index,
            element_tag=target_element.tag,
            element_attrs_key=list(target_element.attrs.keys()),
            element_child_count=child_count,
            element_rect_size=element_rect_size,
            element_rect_center=rect_center,
            is_in_viewport=is_in_viewport,
            is_whole_in_viewport=is_whole_in_viewport,
            is_alive=target_element.states.is_alive,
            is_checked=target_element.states.is_checked,
            is_selected=target_element.states.is_selected,
            is_enabled=target_element.states.is_enabled,
            is_displayed=target_element.states.is_displayed,
            is_covered=is_covered,
            is_clickable=target_element.states.is_clickable,
            has_rect=has_rect
        )


def register_get_screenshot(mcp: FastMCP, browser_manager):
    @mcp.tool(name="get_tab_screenshot",
              description="尝试对传入tab页进行截图，并将截图压缩为1M大小png图片，会返回截图保存路径")
    async def get_tab_screenshot(browser_port: int, tab_id: str) -> dict[str, Any]:
        _browser = browser_manager.get_browser(browser_port)
        target_tab = _browser.get_tab(tab_id)
        target_tab.wait.doc_loaded()
        timestamp = int(time.time() * 1000)
        time.sleep(3)
        origin_png = target_tab.get_screenshot(as_bytes="png")
        compress_png = compress_image_bytes(origin_png)
        image_path = os.path.join(browser_manager.get_cache_path(), f"{browser_port}_{tab_id}_{timestamp}.png")
        with open(image_path, "wb") as f:
            f.write(compress_png)
        return dp_mcp_message_pack(
            message=f"已完成对browser_port={browser_port},tab_id={tab_id}的截屏",
            browser_port=browser_port,
            tab_id=tab_id,
            screenshot_path=image_path
        )



"""
    Drissionpage-mcp中纯html的功能
"""
import os
from typing import Any

from fastmcp import FastMCP

from tools.tools import dp_mcp_message_pack
from DrissionPage.errors import NoRectError


def register_click_action(mcp: FastMCP, browser_manager):
    @mcp.tool(name="click_action",
              description="尝试点击tab页中的元素，返回元素是否可以被点击，以及是否点击成功。"
                          "其中target_element_index默认为0，当传入的Selector可以定位到多个元素时，需要传入target_element_index指定具体点击目标 ")
    async def click_action(browser_port: int, tab_id: str, css_selector: str, target_element_index: int = 0) -> dict[
        str, Any]:
        _browser = browser_manager.get_browser(browser_port)
        target_tab = _browser.get_tab(tab_id)
        css_selector = css_selector
        if "css:" not in css_selector:
            css_selector = "css:" + css_selector
        target_eles = target_tab.eles(css_selector)
        target_element = target_eles[target_element_index]
        element_clickable = target_element.states.is_clickable
        has_rect = False
        is_covered = True
        try:
            has_rect = target_element.states.has_rect
            is_covered = target_element.states.is_covered
            click_success = not is_covered
            if not is_covered:
                click_success = target_element.click()
                click_success = click_success is not False
        except NoRectError:
            has_rect = False
            click_success = False
        except Exception as e:
            click_success = False
        index_message = ""
        if target_element_index > 0:
            index_message = f"【target_element_index={target_element_index}】"
        if element_clickable:
            message = f"tab页:【{tab_id}】点击【Selector={css_selector}】{index_message}的元素 {'成功' if click_success else '失败'} 了"
        else:
            message = f"tab页:【{tab_id}】中【Selector={css_selector}】{index_message}的元素不可被点击"
        if has_rect:
            message += "，元素没有大小和位置信息"
        if has_rect and is_covered:
            message += "，元素被遮挡了"
        return dp_mcp_message_pack(
            message=message,
            browser_port=browser_port,
            tab_id=tab_id,
            css_selector=css_selector,
            element_clickable=element_clickable,
            click_success=click_success,
            extra_message="点击成功，页面可能有更新，请重新获取页面html，并重新分析页面Selector" if element_clickable and click_success else ""
        )


def register_scroll_action(mcp: FastMCP, browser_manager):
    @mcp.tool(name="scroll_action", description="尝试滚动tab页"
                                                "forward参数是滚动的方向：down、up、left、right"
                                                "pixel参数是滚动的像素值，默认为None。"
                                                "当forward为down且pixel为None，则将页面滚动到垂直中间位置，水平位置不变"
                                                "当forward为up且pixel为None，则将页面滚动到顶部，水平位置不变"
                                                "当forward为left且pixel为None，则将页面滚动到最左边，垂直位置不变"
                                                "当forward为right且pixel为None，则将页面滚动到最右边，垂直位置不变")
    async def scroll_action(browser_port: int, tab_id: str, forward: str = "down", pixel: int = None) -> dict[str, Any]:
        _browser = browser_manager.get_browser(browser_port)
        target_tab = _browser.get_tab(tab_id)
        if forward == "down":
            if pixel is None:
                target_tab.scroll.to_half()
            else:
                target_tab.scroll.down(pixel)
        elif forward == "up":
            if pixel is None:
                target_tab.scroll.to_top()
            else:
                target_tab.scroll.up(pixel)
        elif forward == "left":
            if pixel is None:
                target_tab.scroll.to_leftmost()
            else:
                target_tab.scroll.left(pixel)
        elif forward == "right":
            if pixel is None:
                target_tab.scroll.to_rightmost()
            else:
                target_tab.scroll.right(pixel)
        else:
            if pixel is None:
                target_tab.scroll.to_half()
            else:
                target_tab.scroll.down()
        message = f"已完成对tab页:【{tab_id}】forward={forward} 的滑动"
        return dp_mcp_message_pack(
            message=message,
            browser_port=browser_port,
            tab_id=tab_id,
        )


def register_mouse_hover(mcp: FastMCP, browser_manager):
    @mcp.tool(name="mouse_hover", description="将鼠标悬停在元素上【这个功能使用了Drissionpage的action行为链功能】")
    async def mouse_hover(browser_port: int, tab_id: str, css_selector: str, target_element_index: int = 0) -> dict[
        str, Any]:
        _browser = browser_manager.get_browser(browser_port)
        target_tab = _browser.get_tab(tab_id)
        css_selector = css_selector
        if "css:" not in css_selector:
            css_selector = "css:" + css_selector
        target_eles = target_tab.eles(css_selector)
        target_element = target_eles[target_element_index]
        try:
            target_tab.actions.move_to(target_element)
            target_element.hover()
            hover_success = True
        except Exception as e:
            hover_success = False
        if target_element_index > 0:
            message = f"tab页:【{tab_id}】hover【{css_selector}】【index={target_element_index}】的元素 {'成功' if hover_success else '失败'} 了"
        else:
            message = f"tab页:【{tab_id}】hover【{css_selector}】 {'成功' if hover_success else '失败'} 了"
        # else:
        #     message = f"tab页:【{tab_id}】传入的css_selector找到了{len(target_eles)}个元素，请确保传入的css_selector可以找到唯一的一个元素"
        return dp_mcp_message_pack(
            message=message,
            browser_port=browser_port,
            tab_id=tab_id,
            css_selector=css_selector,
            hoversuccess=hover_success,
            extra_message="hover成功，页面可能有更新，请重新获取页面html，并重新分析页面Selector" if hover_success else ""
        )



"""浏览器池管理模块 - 单例模式确保状态共享"""
import random
from typing import Optional, Tuple
import os
from DrissionPage import ChromiumPage, ChromiumOptions
import platform
from DrissionPage.common import Settings

Settings.set_raise_when_click_failed(True)


class BrowserManager:
    """浏览器池管理器 - 使用单例模式"""
    _instance = None

    def __init__(self):
        self.cache_path = os.path.join(os.getcwd(), ".cache")

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.browser_pool = {}
        return cls._instance

    def create_browser(self, user_agent: str = None) -> Tuple[int, ChromiumPage]:
        """创建新的浏览器实例"""
        random_port = random.randint(9223, 9934)
        while random_port in self.browser_pool:
            random_port = random.randint(9223, 9934)

        co = ChromiumOptions().set_local_port(random_port)
        if user_agent:
            co.set_user_agent(user_agent)
        if platform.system() != 'Windows':
            co.set_argument('--no-sandbox')
        custom_data_dir = os.path.join(os.path.expanduser('~'), 'DrissionPage', "userData", f"{random_port}")
        co.set_user_data_path(custom_data_dir)  # 设置用户数据路径
        # if not os.path.exists(custom_data_dir):
        #     os.makedirs(custom_data_dir)
        self.browser_pool[random_port] = ChromiumPage(co)
        return random_port, self.browser_pool[random_port]

    def get_browser(self, port: int) -> Optional[ChromiumPage]:
        """根据端口获取浏览器实例"""
        return self.browser_pool.get(port)

    def remove_browser(self, port: int) -> Tuple[bool, Optional[ChromiumPage]]:
        """根据端口移除浏览器实例"""
        browser = self.browser_pool.pop(port, None)
        return browser is not None, browser

    def list_browsers(self) -> list[int]:
        """列出所有活跃的浏览器端口"""
        return list(self.browser_pool.keys())

    def set_cache_path(self, path: str) -> None:
        self.cache_path = os.path.join(path, ".cache")

    def get_cache_path(self) -> Optional[str]:
        if not os.path.exists(self.cache_path):
            os.makedirs(self.cache_path)
        return self.cache_path


# 创建全局单例实例
browser_manager = BrowserManager()



import threading
from collections import deque
import time
from DrissionPage import ChromiumPage, ChromiumOptions
from DrissionPage._pages.chromium_tab import ChromiumTab
from DrissionPage._units.listener import DataPacket
from typing import Tuple, Optional
import json
from urllib.parse import urlparse, urlunparse

one_turn_max_token = 16000


class DPProxyClient:
    def __init__(self, driver: ChromiumTab, packet_filter: dict, self_kill=False):
        self.tab_id = driver.tab_id
        self.driver = ChromePageProxy(driver, self)
        self.thread = None
        self.self_kill = self_kill
        self.packet_filter = packet_filter
        self.packet_queue = deque()

    def get_driver(self, start_listen, count=None, timeout=10) -> ChromiumTab:
        """
        获取代理后的driver、tab
        :param start_listen: 若你自己写的代码里已经使用自动化框架监听发包的功能了，则该值应该置为False。若没监听，则必须将该值置为True
        :param count: 需捕获的数据包总数，为None表示无限
        :param timeout: 两个数据包之间等待的最大时长（秒），为None表示无限，默认为10秒
        :return:
        """
        if start_listen:
            self.driver.listen.set_targets(res_type=('xhr', 'fetch'))
            self.driver.listen.start()
            self.thread = threading.Thread(target=self.start_listen, args=(count, timeout,))
            self.thread.start()
        return self.driver

    def start_listen(self, count=None, timeout=10):
        for _ in self.driver.listen.steps(count=count, timeout=timeout, gap=1):
            pass

    # 每次调用函数，都从队列的左端弹出一个数据包
    def pop_first_packet(self):
        if self.packet_queue:
            result = self.packet_queue.popleft()
            current_queue_size = len(self.packet_queue)
            return current_queue_size, json.dumps(result, ensure_ascii=False, separators=(',', ':')).replace("\\", "")
        else:
            return 0, None


class DPProxyClientManager:
    """浏览器池管理器 - 使用单例模式"""
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.tab_pool = {}
        return cls._instance

    def create_client(self, tab: ChromiumTab, packet_filter: dict, self_kill=False) -> Tuple[
        str, DPProxyClient, ChromiumTab]:
        """创建新的tab页面代理实例"""
        client = DPProxyClient(tab, packet_filter, self_kill=self_kill)
        tab = client.get_driver(True, timeout=60 * 10)
        tab_id = tab.tab_id
        self.tab_pool[tab_id] = {"client": client, "driver": tab}
        return tab_id, client, tab

    def get_client(self, tab_id: str) -> Optional[DPProxyClient]:
        """根据端口获取浏览器实例"""
        return self.tab_pool.get(tab_id).get("client", None)

    def remove_client(self, tab_id: str) -> Tuple[bool, Optional[ChromiumPage]]:
        """根据端口移除浏览器实例"""
        client = self.tab_pool.pop(tab_id, None)
        return client is not None, client

    def list_clients(self) -> list[int]:
        """列出所有活跃的浏览器端口"""
        return list(self.tab_pool.keys())


class ChromePageProxy:
    def __init__(self, page, client=None):
        self.__dict__['page'] = page
        self.__dict__['client'] = client

    def __getattr__(self, item):
        attr = getattr(self.page, item)
        print(item, attr)
        if item == 'listen':
            listen_proxy = DrissionPageListenerProxy(attr, self.__dict__['client'])
            return listen_proxy
        return attr


class DrissionPageListenerProxy:
    def __init__(self, listener, client=None):
        self.listener = listener
        self.client = client

    def __getattr__(self, item):
        attr = getattr(self.listener, item)
        # 当监听到wait被调用的时候
        if item == "wait":
            def wrapper(*args, **kwargs):
                result = attr(*args, **kwargs)
                check_data_packet(result, self.client)
                return result

            return wrapper
        # 当监听到steps被调用的时候
        if item == "steps":
            def wrapper(*args, **kwargs):
                if kwargs.get("gap", 1) > 1:
                    raise Exception("暂不支持多包监控")
                result = attr(*args, **kwargs)
                if attr.__name__ == "steps":
                    for step in result:
                        check_data_packet(step, self.client)
                        yield step

            return wrapper
        return attr


def check_data_packet(packet: DataPacket, client: DPProxyClient):
    """
    封装监听到的数据包，并将其存放在client的packet_queue中
    :param packet:
    :param client:
    :return:
    """
    url = packet.url
    method = packet.request.method
    data = None
    if packet.request.hasPostData:
        data = packet.request.postData
    domain = urlparse(url).netloc
    body = packet.response.body
    if isinstance(body, dict):
        body_str = json.dumps(body, ensure_ascii=False, separators=(',', ':'))
    else:
        body_str = str(body)
    body_str_list = [body_str[i:i + one_turn_max_token] for i in range(0, len(body_str), one_turn_max_token)]
    body_completed = True
    packet_filter = client.packet_filter
    domain_filter = packet_filter.get("domain_filter", None)
    method_filter = packet_filter.get("method_filter", ["GET", "POST"])
    for index, body_str in enumerate(body_str_list):
        # 如果给了domain_filter并且domain没有在domain_filter中时跳过该数据包
        if domain_filter and domain not in domain_filter:
            continue
        # 如果method没有在method_filter中，则跳过该数据包
        if method not in method_filter:
            continue
        if (index + 1) != len(body_str_list):
            body_completed = False
        try:
            response_headers = packet.response.headers
        except TypeError:
            response_headers = {}
        temp_dict = {
            "url": url,
            "body_completed": body_completed,
            "method": method,
            "request_data": data,
            "request_headers": dict(packet.request.headers),
            "response_headers": dict(response_headers),
            "response_body_segment": body_str.replace("\\", ""),
        }
        client.packet_queue.append(temp_dict)


client_manager = DPProxyClientManager()



"""
    Drissionpage-mcp中纯Cv领域的相关功能，为Vl模型赋能
"""
from typing import Any, Tuple

from fastmcp import FastMCP
from typing import Annotated, Dict, List
from tools.tools import dp_mcp_message_pack, compress_html
from DrissionPage.errors import NoRectError
import os

ele_outer_html_max_len = 500


def register_get_ele_info_by_bboxes(mcp: FastMCP, browser_manager):
    @mcp.tool(name="get_ele_info_by_bboxes",
              description="根据传入的bboxes字典返回，对应的元素信息列表")
    async def get_ele_info_by_bboxes(
            browser_port: Annotated[int, "浏览器端口"],
            tab_id: Annotated[str, "tab页id"],
            element_bboxes_list: Annotated[Dict[
                str, list], "元素bboxes字典，字典格式为：{'element_name':[x1, y1, x2, y2]}，左上 (x1,y1) → 右下 (x2,y2)，其中每个x和y都是(0,1)之间的float值"]
    ) -> dict[str, Any]:
        _browser = browser_manager.get_browser(browser_port)
        target_tab = _browser.get_tab(tab_id)
        viewport_size_with_scrollbar = target_tab.rect.viewport_size_with_scrollbar
        result_dict = dict()
        for element_name, bboxes in element_bboxes_list.items():
            x_center = (bboxes[0] + bboxes[2]) / 2 * viewport_size_with_scrollbar[0]
            y_center = (bboxes[1] + bboxes[3]) / 2 * viewport_size_with_scrollbar[1]
            target_ele = target_tab.run_js(f'return document.elementFromPoint({x_center}, {y_center})')
            temp_dict = {
                "bboxes圈定的中心": (x_center, y_center),
                "当前窗口大小": viewport_size_with_scrollbar,
                "bboxes": bboxes,
            }
            if target_ele is not None:
                target_ele = target_tab.ele(target_ele)
                try:
                    has_rect = target_ele.states.has_rect
                    is_covered = target_ele.states.is_covered
                    rect_x_center = (has_rect[0][0] + has_rect[1][0]) / 2
                    rect_y_center = (has_rect[1][1] + has_rect[2][1]) / 2
                    rect_center = (rect_x_center, rect_y_center)
                    rect_size = target_ele.rect.size
                except NoRectError:
                    has_rect = False
                    is_covered = "元素没有大小和位置信息"
                    rect_center = "元素没有大小和位置信息"
                    rect_size = "元素没有大小和位置信息"
                temp_dict["是否定位到元素"] = True
                compressed_html, rate = compress_html(target_ele.html)
                if len(compressed_html) > ele_outer_html_max_len:
                    file_name = f"{tab_id}_{x_center}_{y_center}_segment.html"
                    abs_path = os.path.join(browser_manager.get_cache_path(), file_name)
                    with open(abs_path, "w", encoding="utf-8") as f:
                        f.write(compressed_html)
                    compressed_html = f"元素的outerHTML过大，已将元素的outerHTML保存至：{abs_path}"
                temp_dict["元素信息"] = {
                    "元素tag": target_ele.tag,
                    "元素在页面中 css selector 的绝对路径": target_ele.css_path,
                    "元素是否拥有大小和位置信息": has_rect,
                    "元素rect确定的中心": rect_center,
                    "元素rect的大小": rect_size,
                    "元素是否可被模拟点击": target_ele.states.is_clickable,
                    "元素是否可见": target_ele.states.is_displayed,
                    "元素是否可用": target_ele.states.is_enabled,
                    "元素是否被其它元素覆盖": is_covered,
                    "元素的outerHTML": compressed_html,
                }
            else:
                temp_dict["是否定位到元素"] = False
                temp_dict["元素信息"] = {}
            result_dict[element_name] = temp_dict
        message = f"对于【browser_port={browser_port}】【tab_id={tab_id}】成功选中{len(result_dict)}个元素。"
        return dp_mcp_message_pack(
            message=message,
            browser_port=browser_port,
            tab_id=tab_id,
            element_info=result_dict,
        )


def register_click_action_by_xy(mcp: FastMCP, browser_manager):
    @mcp.tool(name="click_action_by_xy",
              description="根据传入的坐标，进行点击，返回是否点击成功。")
    async def get_ele_info_by_bboxes(
            browser_port: Annotated[int, "浏览器端口"],
            tab_id: Annotated[str, "tab页id"],
            coordinates: Annotated[Tuple, "要点击的坐标，这个坐标不可以是bboxes"],
            relocate_element: Annotated[
                bool, "是否要在这个工具内使用coordinates参数重新定位元素，这可能导致点击的元素与预期的不一致"] = False,
    ) -> dict[str, Any]:
        _browser = browser_manager.get_browser(browser_port)
        target_tab = _browser.get_tab(tab_id)
        if relocate_element:
            target_ele = target_tab.run_js(f'return document.elementFromPoint({coordinates[0]}, {coordinates[1]})')
            if target_ele is None:
                message = f"使用【coordinates={coordinates}】未能重新定位到元素"
                return dp_mcp_message_pack(
                    message=message,
                    browser_port=browser_port,
                    tab_id=tab_id,
                )
            target_element = target_tab.ele(target_ele)
            element_clickable = target_element.states.is_clickable
            has_rect = False
            is_covered = True
            rect_center = None
            try:
                has_rect = target_element.states.has_rect
                is_covered = target_element.states.is_covered
                if has_rect:
                    rect_x_center = (has_rect[0][0] + has_rect[1][0]) / 2
                    rect_y_center = (has_rect[1][1] + has_rect[2][1]) / 2
                    rect_center = (rect_x_center, rect_y_center)
                click_success = not is_covered
                if not is_covered:
                    click_success = target_element.click()
                    click_success = click_success is not False
            except NoRectError:
                has_rect = False
                click_success = False
            except Exception as e:
                click_success = False
            rect_center_message = ""
            if rect_center:
                rect_center_message = f"，实际rect_center={rect_center}"
            if element_clickable:
                message = f"tab页:【{tab_id}】点击【coordinates={coordinates}{rect_center_message}】的元素 {'成功' if click_success else '失败'} 了"
            else:
                message = f"tab页:【{tab_id}】中【coordinates={coordinates}{rect_center_message}】的元素不可被点击"
            if has_rect:
                message += "，元素没有大小和位置信息"
            if has_rect and is_covered:
                message += "，元素被遮挡了"
            return dp_mcp_message_pack(
                message=message,
                browser_port=browser_port,
                tab_id=tab_id,
                css_selector=target_element.css_path,
                element_clickable=element_clickable,
                click_success=click_success,
                extra_message="点击成功，页面可能有更新，请重新获取页面html，并重新分析页面Selector" if element_clickable and click_success else ""
            )
        try:
            target_tab.actions.move_to(coordinates)
            click_success = target_tab.actions.click()
            click_success = click_success is not False
        except Exception as e:
            click_success = False
        message=f"tab页:【{tab_id}】中【coordinates={coordinates}】位置点击 {'成功' if click_success else '失败'} 了"
        return dp_mcp_message_pack(
            message=message,
            browser_port=browser_port,
            tab_id=tab_id,
            click_success=click_success,

        )



from fastmcp import FastMCP

from mcp_tools.dp_base import *
from mcp_tools.dp_html import *
from mcp_tools.dp_cv import *
from tools.browser_manager import browser_manager
from tools.browser_proxy import client_manager

mcp = FastMCP("Jarvis Brain Mcp Tools")

# 根据环境变量加载模块
enabled_modules = os.getenv("MCP_MODULES", "Dp-Base,Dp-Html").split(",")
base_cwd = os.getenv("BASE_CWD", os.path.expanduser('~'))
browser_manager.set_cache_path(base_cwd)

if "Dp-Base" in enabled_modules:
    # 页面管理
    register_close_tab(mcp, browser_manager)
    register_switch_tab(mcp, browser_manager)
    register_get_new_tab(mcp, browser_manager, client_manager)
    register_quit_browser(mcp, browser_manager)
    # 基础功能
    register_visit_url(mcp, browser_manager, client_manager)
    register_get_html(mcp, browser_manager)
    register_check_selector(mcp, browser_manager)
    register_pop_first_packet(mcp, browser_manager, client_manager)
    register_get_ele_info(mcp, browser_manager)
    # 页面交互
    register_scroll_action(mcp, browser_manager)
    register_get_screenshot(mcp, browser_manager)

if "Dp-Html" in enabled_modules:
    # 使用html分析出来的Selector进行交互
    # 页面交互
    register_click_action(mcp, browser_manager)
    register_mouse_hover(mcp, browser_manager)

if "Dp-Cv" in enabled_modules:
    # 使用Vl模型分析出来的坐标进行交互
    # 页面交互
    register_get_ele_info_by_bboxes(mcp, browser_manager)
    register_click_action_by_xy(mcp, browser_manager)


def main():
    mcp.run(transport="stdio", show_banner=False)


if __name__ == '__main__':
    main()
